@@ -1,5 +1,5 @@
 # emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
-# ex: set sts=4 ts=4 sw=4 noet:
+# ex: set sts=4 ts=4 sw=4 et:
 # ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
 #
 #   See COPYING file distributed along with the datalad package for the
@@ -193,6 +193,7 @@ class Create(Interface):
     def __call__(
             path=None,
             initopts=None,
+            *,
             force=False,
             description=None,
             dataset=None,
@@ -339,6 +340,26 @@ class Create(Interface):
             yield res
             return
 
+        # Check if specified cfg_proc(s) can be discovered, storing
+        # the results so they can be used when the time comes to run
+        # the procedure. If a procedure cannot be found, raise an
+        # error to prevent creating the dataset.
+        cfg_proc_specs = []
+        if cfg_proc:
+            discovered_procs = tbds.run_procedure(
+                discover=True,
+                result_renderer='disabled',
+                return_type='generator',
+            )
+            for cfg_proc_ in cfg_proc:
+                for discovered_proc in discovered_procs:
+                    if discovered_proc['procedure_name'] == 'cfg_' + cfg_proc_:
+                        cfg_proc_specs.append(discovered_proc)
+                        break
+                else:
+                    raise ValueError("Cannot find procedure with name "
+                                     "'%s'" % cfg_proc_)
+
         if initopts is not None and isinstance(initopts, list):
             initopts = {'_from_cmdline_': initopts}
 
@@ -367,7 +388,7 @@ class Create(Interface):
         if id_var in tbds_config:
             # make sure we reset this variable completely, in case of a
             # re-create
-            tbds_config.unset(id_var, where='dataset')
+            tbds_config.unset(id_var, scope='branch')
 
         if _seed is None:
             # just the standard way
@@ -379,7 +400,7 @@ class Create(Interface):
         tbds_config.add(
             id_var,
             tbds_id if tbds_id is not None else uuid_id,
-            where='dataset',
+            scope='branch',
             reload=False)
 
         # make config overrides permanent in the repo config
@@ -389,7 +410,7 @@ class Create(Interface):
         # and unnecessary for the Python API (there could simply be a
         # subsequence ds.config.add() call)
         for k, v in tbds_config.overrides.items():
-            tbds_config.add(k, v, where='local', reload=False)
+            tbds_config.add(k, v, scope='local', reload=False)
 
         # all config manipulation is done -> fll reload
         tbds_config.reload()
@@ -411,9 +432,12 @@ class Create(Interface):
             _status=add_to_git,
         )
 
-        for cfg_proc_ in cfg_proc:
-            for r in tbds.run_procedure('cfg_' + cfg_proc_):
-                yield r
+        for cfg_proc_spec in cfg_proc_specs:
+            yield from tbds.run_procedure(
+                cfg_proc_spec,
+                result_renderer='disabled',
+                return_type='generator',
+            )
 
         # the next only makes sense if we saved the created dataset,
         # otherwise we have no committed state to be registered
@@ -421,10 +445,11 @@ class Create(Interface):
         if isinstance(refds, Dataset) and refds.path != tbds.path:
             # we created a dataset in another dataset
             # -> make submodule
-            for r in refds.save(
-                    path=tbds.path,
-            ):
-                yield r
+            yield from refds.save(
+                path=tbds.path,
+                return_type='generator',
+                result_renderer='disabled',
+            )
 
         res.update({'status': 'ok'})
         yield res
