@@ -1,5 +1,5 @@
 # emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
-# ex: set sts=4 ts=4 sw=4 noet:
+# ex: set sts=4 ts=4 sw=4 et:
 # ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
 #
 #   See COPYING file distributed along with the datalad package for the
@@ -12,8 +12,8 @@ __docformat__ = 'restructuredtext'
 
 
 import logging
-import subprocess
 
+from datalad.cmd import WitlessRunner as Runner
 from datalad.interface.common_opts import (
     recursion_flag,
     recursion_limit
@@ -41,6 +41,8 @@ from datalad.distribution.dataset import (
 )
 from datalad.distributed.ora_remote import (
     LocalIO,
+    RIARemoteError,
+    RemoteCommandFailedError,
     SSHRemoteIO,
 )
 from datalad.utils import (
@@ -146,10 +148,22 @@ class CreateSiblingRia(Interface):
             constraints=EnsureDataset() | EnsureNone()),
         url=Parameter(
             args=("url",),
-            metavar="ria+<ssh|file>://<host>[/path]",
-            doc="""URL identifying the target RIA store and access protocol.
+            metavar="ria+<ssh|file|http(s)>://<host>[/path]",
+            doc="""URL identifying the target RIA store and access protocol. If
+            ``push_url||--push-url`` is given in addition, this is
+            used for read access only. Otherwise it will be used for write
+            access too and to create the repository sibling in the RIA store.
+            Note, that HTTP(S) currently is valid for consumption only thus
+            requiring to provide ``push_url||--push-url``.
             """,
             constraints=EnsureStr() | EnsureNone()),
+        push_url=Parameter(
+            args=("--push-url",),
+            metavar="ria+<ssh|file>://<host>[/path]",
+            doc="""URL identifying the target RIA store and access protocol for
+            write access to the storage sibling. If given this will also be used
+            for creation of the repository sibling in the RIA store.""",
+            constraints=EnsureStr() | EnsureNone()),
         name=Parameter(
             args=('-s', '--name',),
             metavar='NAME',
@@ -167,6 +181,14 @@ class CreateSiblingRia(Interface):
             a storage sibling is created, this setting is ignored, and
             the primary sibling name is used.""",
             constraints=EnsureStr() | EnsureNone()),
+        alias=Parameter(
+            args=('--alias',),
+            metavar='ALIAS',
+            doc="""Alias for the dataset in the RIA store.
+            Add the necessary symlink so that this dataset can be cloned from the RIA
+            store using the given ALIAS instead of its ID.
+            With `recursive=True`, only the top dataset will be aliased.""",
+            constraints=EnsureStr() | EnsureNone()),
         post_update_hook=Parameter(
             args=("--post-update-hook",),
             doc="""Enable git's default post-update-hook for the created
@@ -200,8 +222,7 @@ class CreateSiblingRia(Interface):
             installation is required on the target host."""),
         existing=Parameter(
             args=("--existing",),
-            constraints=EnsureChoice(
-                'skip', 'error', 'reconfigure') | EnsureNone(),
+            constraints=EnsureChoice('skip', 'error', 'reconfigure', None),
             metavar='MODE',
             doc="""Action to perform, if a (storage) sibling is already
             configured under the given name and/or a target already exists.
@@ -209,13 +230,18 @@ class CreateSiblingRia(Interface):
             repository be forcefully re-initialized, and the sibling
             (re-)configured ('reconfigure'), or the command be instructed to
             fail ('error').""", ),
+        new_store_ok=Parameter(
+            args=("--new-store-ok",),
+            action='store_true',
+            doc="""When set, a new store will be created, if necessary. Otherwise, a sibling
+            will only be created if the url points to an existing RIA store.""",
+        ),
         recursive=recursion_flag,
         recursion_limit=recursion_limit,
         trust_level=Parameter(
             args=("--trust-level",),
             metavar="TRUST-LEVEL",
-            constraints=EnsureChoice(
-                'trust', 'semitrust', 'untrust') | EnsureNone(),
+            constraints=EnsureChoice('trust', 'semitrust', 'untrust', None),
             doc="""specify a trust level for the storage sibling. If not
             specified, the default git-annex trust level is used. 'trust'
             should be used with care (see the git-annex-trust man page).""",),
@@ -232,17 +258,21 @@ class CreateSiblingRia(Interface):
     @eval_results
     def __call__(url,
                  name,
+                 *,  # note that `name` is required but not posarg in CLI
                  dataset=None,
                  storage_name=None,
+                 alias=None,
                  post_update_hook=False,
                  shared=None,
                  group=None,
                  storage_sibling=True,
                  existing='error',
+                 new_store_ok=False,
                  trust_level=None,
                  recursive=False,
                  recursion_limit=None,
                  disable_storage__=None,
+                 push_url=None
                  ):
         if disable_storage__ is not None:
             import warnings
@@ -269,8 +299,12 @@ class CreateSiblingRia(Interface):
         )
 
         # parse target URL
+        # Note: URL parsing is done twice ATM (for top-level ds). This can't be
+        # reduced to single instance, since rewriting url based on config could
+        # be different for subdatasets.
         try:
-            ssh_host, base_path, rewritten_url = verify_ria_url(url, ds.config)
+            ssh_host, base_path, rewritten_url = \
+                verify_ria_url(push_url if push_url else url, ds.config)
         except ValueError as e:
             yield get_status_dict(
                 status='error',
@@ -321,7 +355,8 @@ class CreateSiblingRia(Interface):
             # even if we have to fail, let's report all conflicting siblings
             # in subdatasets
             failed = False
-            for r in ds.siblings(result_renderer=None,
+            for r in ds.siblings(result_renderer='disabled',
+                                 return_type='generator',
                                  recursive=recursive,
                                  recursion_limit=recursion_limit):
                 log_progress(
@@ -367,20 +402,41 @@ class CreateSiblingRia(Interface):
         #         command abstractions
         #       - more generally consider store creation a dedicated command or
         #         option
-        # Note: URL parsing is done twice ATM (for top-level ds). This can't be
-        # reduced to single instance, since rewriting url based on config could
-        # be different for subdatasets.
 
-        create_store(SSHRemoteIO(ssh_host) if ssh_host else LocalIO(),
+        io = SSHRemoteIO(ssh_host) if ssh_host else LocalIO()
+        try:
+            # determine the existence of a store by trying to read its layout.
+            # Because this raises a FileNotFound error if non-existent, we need
+            # to catch it
+            io.read_file(Path(base_path) / 'ria-layout-version')
+        except (FileNotFoundError, RIARemoteError, RemoteCommandFailedError) as e:
+            if not new_store_ok:
+                # we're instructed to only act in case of an existing RIA store
+                res = get_status_dict(
+                    status='error',
+                    message="No store found at '{}'. Forgot "
+                            "--new-store-ok ?".format(
+                        Path(base_path)),
+                    **res_kwargs)
+                yield res
+                return
+
+        log_progress(
+            lgr.info, 'create-sibling-ria',
+            'Creating a new RIA store at %s', Path(base_path),
+        )
+        create_store(io,
                      Path(base_path),
                      '1')
 
         yield from _create_sibling_ria(
             ds,
             url,
+            push_url,
             name,
             storage_sibling,
             storage_name,
+            alias,
             existing,
             shared,
             group,
@@ -393,16 +449,20 @@ class CreateSiblingRia(Interface):
             # recursion when querying for them and _no_recursion with the
             # actual call. Theoretically this can be parallelized.
 
-            for subds in ds.subdatasets(fulfilled=True,
+            for subds in ds.subdatasets(state='present',
                                         recursive=True,
                                         recursion_limit=recursion_limit,
+                                        return_type='generator',
+                                        result_renderer='disabled',
                                         result_xfm='datasets'):
                 yield from _create_sibling_ria(
                     subds,
                     url,
+                    push_url,
                     name,
                     storage_sibling,
                     storage_name,
+                    None,  # subdatasets can't have the same alias as the parent
                     existing,
                     shared,
                     group,
@@ -414,9 +474,11 @@ class CreateSiblingRia(Interface):
 def _create_sibling_ria(
         ds,
         url,
+        push_url,
         name,
         storage_sibling,
         storage_name,
+        alias,
         existing,
         shared,
         group,
@@ -438,7 +500,8 @@ def _create_sibling_ria(
 
     # parse target URL
     try:
-        ssh_host, base_path, rewritten_url = verify_ria_url(url, ds.config)
+        ssh_host, base_path, rewritten_url = \
+            verify_ria_url(push_url if push_url else url, ds.config)
     except ValueError as e:
         yield get_status_dict(
             status='error',
@@ -454,10 +517,19 @@ def _create_sibling_ria(
         url + '#{}'.format(ds.id),
         cfg=ds.config
     )['giturl']
-    # determine layout locations; go for a v1 layout
+    git_push_url = decode_source_spec(
+        push_url + '#{}'.format(ds.id),
+        cfg=ds.config
+    )['giturl'] if push_url else None
+
+    # determine layout locations; go for a v1 store-level layout
     repo_path, _, _ = get_layout_locations(1, base_path, ds.id)
 
-    ds_siblings = [r['name'] for r in ds.siblings(result_renderer=None)]
+    ds_siblings = [
+        r['name'] for r in ds.siblings(
+            result_renderer='disabled',
+            return_type='generator')
+    ]
     # Figure whether we are supposed to skip this very dataset
     if existing == 'skip' and (
             name in ds_siblings or (
@@ -525,7 +597,8 @@ def _create_sibling_ria(
             " and '{}'".format(storage_name) if storage_name else '',
         ))
     create_ds_in_store(SSHRemoteIO(ssh_host) if ssh_host else LocalIO(),
-                       base_path, ds.id, '2', '1')
+                       base_path, ds.id, '2', '1', alias,
+                       init_obj_tree=storage_sibling is not False)
     if storage_sibling:
         # we are using the main `name`, if the only thing we are creating
         # is the storage sibling
@@ -538,6 +611,8 @@ def _create_sibling_ria(
             'encryption=none',
             'autoenable=true',
             'url={}'.format(url)]
+        if push_url:
+            special_remote_options.append('push-url={}'.format(push_url))
         try:
             ds.repo.init_remote(
                 srname,
@@ -637,13 +712,15 @@ def _create_sibling_ria(
             # write special remote's uuid into git-config, so clone can
             # which one it is supposed to be and enable it even with
             # fallback URL
-            gr.config.add("datalad.ora-remote.uuid", uuid, where='local')
+            gr.config.add("datalad.ora-remote.uuid", uuid, scope='local')
 
         if post_update_hook:
             disabled_hook.rename(enabled_hook)
         if group:
-            # TODO; do we need a cwd here?
-            subprocess.run(chgrp_cmd, cwd=ds.path)
+            # No CWD needed here, since `chgrp` is expected to be found via PATH
+            # and the path it's operating on is absolute (repo_path). No
+            # repository operation involved.
+            Runner().run(chgrp_cmd)
         # finally update server
         if post_update_hook:
             # Conditional on post_update_hook, since one w/o the other doesn't
@@ -663,17 +740,17 @@ def _create_sibling_ria(
     ds.config.set(
         "remote.{}.annex-ignore".format(name),
         value="true",
-        where="local")
-    ds.siblings(
+        scope="local")
+    yield from ds.siblings(
         'configure',
         name=name,
-        url=git_url
-        if ssh_host
-        else str(repo_path),
+        url=str(repo_path) if url.startswith("ria+file") else git_url,
+        pushurl=git_push_url,
         recursive=False,
         # Note, that this should be None if storage_sibling was not set
         publish_depends=storage_name,
-        result_renderer=None,
+        result_renderer='disabled',
+        return_type='generator',
         # Note, that otherwise a subsequent publish will report
         # "notneeded".
         fetch=True
