@@ -1,6 +1,6 @@
 # emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
 # -*- coding: utf-8 -*-
-# ex: set sts=4 ts=4 sw=4 noet:
+# ex: set sts=4 ts=4 sw=4 et:
 # ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
 #
 #   See COPYING file distributed along with the datalad package for the
@@ -11,13 +11,16 @@
 
 """
 
+from functools import wraps
+import inspect
 import os
 import os.path as op
 import shutil
+import stat
 import sys
+import time
 import logging
 from unittest.mock import patch
-import builtins
 
 from operator import itemgetter
 from os.path import (
@@ -38,27 +41,31 @@ from datalad.utils import (
     _path_,
     all_same,
     any_re_search,
-    assure_unicode,
     auto_repr,
     better_wraps,
     CMD_MAX_ARG,
+    Path,
     create_tree,
     disable_logger,
     dlabspath,
+    ensure_write_permission,
     expandpath,
     file_basename,
     find_files,
     generate_chunks,
+    getargspec,
     get_dataset_root,
-    get_func_kwargs_doc,
     get_open_files,
     get_path_prefix,
+    get_sig_param_names,
     get_timestamp_suffix,
     get_trace,
     getpwd, chpwd,
     import_module_from_file,
     import_modules,
     is_explicit_path,
+    is_interactive,
+    join_cmdline,
     knows_annex,
     line_profile,
     make_tempfile,
@@ -66,14 +73,13 @@ from datalad.utils import (
     md5sum,
     never_fail,
     not_supported_on_windows,
+    obtain_write_permission,
     on_windows,
     partition,
-    Path,
     path_is_subpath,
     path_startswith,
     rotree,
-    safe_print,
-    setup_exceptionhook,
+    split_cmdline,
     swallow_logs,
     swallow_outputs,
     unique,
@@ -92,28 +98,27 @@ from .utils import (
     assert_not_in,
     assert_raises,
     assert_true,
-    assure_bool,
-    assure_dict_from_str,
-    assure_iter,
-    assure_list,
-    assure_list_from_str,
-    assure_unicode,
+    ensure_bool,
+    ensure_dict_from_str,
+    ensure_iter,
+    ensure_list,
+    ensure_list_from_str,
+    ensure_unicode,
     eq_,
     has_symlink_capability,
     known_failure,
-    known_failure_appveyor,
-    known_failure_v6,
-    known_failure_windows,
     nok_,
     OBSCURE_FILENAME,
     ok_,
     ok_file_has_content,
     ok_generator,
     ok_startswith,
+    on_travis,
     probe_known_failure,
     skip_if,
     skip_if_no_module,
     skip_if_on_windows,
+    skip_if_root,
     skip_known_failure,
     SkipTest,
     skip_wo_symlink_capability,
@@ -123,16 +128,7 @@ from .utils import (
 from datalad import cfg as dl_cfg
 
 
-def test_get_func_kwargs_doc():
-    def some_func(arg1, kwarg1=None, kwarg2="bu"):
-        return
-    eq_(get_func_kwargs_doc(some_func), ['arg1', 'kwarg1', 'kwarg2'])
-
-
 def test_better_wraps():
-    from functools import wraps
-    from datalad.utils import getargspec
-
     def wraps_decorator(func):
         @wraps(func)
         def  _wrap_wraps_decorator(*args, **kwargs):
@@ -156,11 +152,80 @@ def test_better_wraps():
         return "function2"
 
     eq_("function1", function1(1, 2, 3))
-    eq_(getargspec(function1)[0], [])
+    # getargspec shim now can handle @wraps'ed functions just fine
+    eq_(getargspec(function1)[0], ['a', 'b', 'c'])
     eq_("function2", function2(1, 2, 3))
     eq_(getargspec(function2)[0], ['a', 'b', 'c'])
 
 
+def test_getargspec():
+
+    def eq_argspec(f, expected, has_kwonlyargs=False):
+        """A helper to centralize testing of getargspec on original and wrapped function
+
+        has_kwonlyargs is to instruct if function has kwonly args so we do not try to compare
+        to inspect.get*spec functions, which would barf ValueError if attempted to run on a
+        function with kwonlys. And also we pass it as include_kwonlyargs to our getargspec
+        """
+        # so we know that our expected is correct
+        if not has_kwonlyargs:
+            # if False - we test function with kwonlys - inspect.getargspec would barf
+            eq_(inspect.getargspec(f), expected)
+            # and getfullargspec[:4] wouldn't provide a full picture
+            eq_(inspect.getfullargspec(f)[:4], expected)
+        else:
+            assert_raises(ValueError, inspect.getargspec, f)
+            inspect.getfullargspec(f)  # doesn't barf
+        eq_(getargspec(f, include_kwonlyargs=has_kwonlyargs), expected)
+
+        # and lets try on a wrapped one -- only ours can do the right thing
+        def decorator(f):
+            @wraps(f)
+            def wrapper(*args, **kwargs):  # pragma: no cover
+                return f(*args, **kwargs)
+            return wrapper
+        fw = decorator(f)
+        if has_kwonlyargs:
+            # We barf ValueError similarly to inspect.getargspec, unless explicitly requested
+            # to include kwonlyargs
+            assert_raises(ValueError, getargspec, fw)
+        eq_(getargspec(fw, include_kwonlyargs=has_kwonlyargs), expected)
+
+    def f0():  # pragma: no cover
+        pass
+
+    yield eq_argspec, f0, ([], None, None, None)
+
+    def f1(a1, kw1=None, kw0=1):  # pragma: no cover
+        pass
+
+    yield eq_argspec, f1, (['a1', 'kw1', 'kw0'], None, None, (None, 1))
+
+    # Having *a already makes keyword args to be kwonlyargs, in that
+    # inspect.get*spec would barf
+    def f1_args(a1, *a, kw1=None, kw0=1, **kw):  # pragma: no cover
+        pass
+
+    yield eq_argspec, f1_args, (['a1', 'kw1', 'kw0'], 'a', 'kw', (None, 1)), True
+
+    def f1_star(a1, *, kw1=None, kw0=1):  # pragma: no cover
+        pass
+
+    assert_raises(ValueError, getargspec, f1_star)
+    yield eq_argspec, f1_star, (['a1', 'kw1', 'kw0'], None, None, (None, 1)), True
+
+
+def test_get_sig_param_names():
+    def f(a1, kw1=None, *args, kw2=None, **kwargs):
+        pass  # pragma: no cover
+
+    # note: `a1` could be used either positionally or via keyword, so is listed in kw_any
+    assert_equal(get_sig_param_names(f, ('kw_only', 'kw_any')), (['kw2'], ['a1', 'kw1', 'kw2']))
+    assert_equal(get_sig_param_names(f, ('any',)), (['a1', 'kw1', 'kw2'],))
+    assert_equal(get_sig_param_names(f, tuple()), ())
+    assert_raises(ValueError, get_sig_param_names, f, ('mumba',))
+
+
 @with_tempfile(mkdir=True)
 def test_rotree(d):
     d2 = opj(d, 'd1', 'd2')  # deep nested directory
@@ -281,51 +346,6 @@ def test_disable_logger():
         assert_raises(AssertionError, cml.assert_logged, "log sth at bottom level")
 
 
-def _check_setup_exceptionhook(interactive):
-    old_exceptionhook = sys.excepthook
-
-    post_mortem_tb = []
-
-    def our_post_mortem(tb):
-        post_mortem_tb.append(tb)
-
-    with patch('sys.excepthook'), \
-            patch('datalad.utils.is_interactive', lambda: interactive), \
-            patch('pdb.post_mortem', our_post_mortem):
-        setup_exceptionhook()
-        our_exceptionhook = sys.excepthook
-        ok_(old_exceptionhook != our_exceptionhook)
-        with swallow_logs() as cml, swallow_outputs() as cmo:
-            # we need to call our_exceptionhook explicitly b/c nose
-            # swallows all Exceptions and hook never gets executed
-            try:
-                raise RuntimeError
-            except Exception as e:  # RuntimeError:
-                type_, value_, tb_ = sys.exc_info()
-            our_exceptionhook(type_, value_, tb_)
-            # Happens under tox environment but not in manually crafted
-            # ones -- not yet sure what it is about but --dbg does work
-            # with python3 so lettting it skip for now
-            raise SkipTest(
-                "TODO: Not clear why in PY3 calls cleanup if we try to "
-                "access the beast"
-            )
-            assert_in('Traceback (most recent call last)', cmo.err)
-            assert_in('in _check_setup_exceptionhook', cmo.err)
-            if interactive:
-                assert_equal(post_mortem_tb[0], tb_)
-            else:
-                assert_equal(post_mortem_tb, [])
-                # assert_in('We cannot setup exception hook', cml.out)
-
-    eq_(old_exceptionhook, sys.excepthook)
-
-
-def test_setup_exceptionhook():
-    for tval in [True, False]:
-        yield _check_setup_exceptionhook, tval
-
-
 def test_md5sum():
     # just a smoke (encoding/decoding) test for md5sum
     _ = md5sum(__file__)
@@ -370,8 +390,8 @@ def test_getpwd_basic():
         assert_false(oschdir.called)
 
 
-@assert_cwd_unchanged(ok_to_chdir=True)
 @with_tempfile(mkdir=True)
+@assert_cwd_unchanged(ok_to_chdir=True)
 def test_getpwd_change_mode(tdir):
     from datalad import utils
     if utils._pwd_mode != 'PWD':
@@ -443,55 +463,70 @@ def test_auto_repr():
     class buga:
         def __init__(self):
             self.a = 1
-            self.b = list(range(100))
+            self.b = list(range(20))
             self.c = WithoutReprClass()
             self._c = "protect me"
 
         def some(self):
             return "some"
 
+    @auto_repr(short=False)
+    class buga_long(object):
+        def __init__(self):
+            self.a = 1
+            self.b = list(range(20))
+
+        def some(self):
+            return "some"
+
     assert_equal(
         repr(buga()),
-        "buga(a=1, b=<<[0, 1, 2, 3, 4++372 chars++ 99]>>, c=<WithoutReprClass>)"
+        "buga(a=1, b=<<[0, 1, 2, 3, 4++52 chars++ 19]>>, c=<WithoutReprClass>)"
     )
     assert_equal(buga().some(), "some")
 
+    assert_equal(
+        repr(buga_long()),
+        f"buga_long(a=1, b=[{', '.join(map(str, range(20)))}])"
+    )
+    assert_equal(buga_long().some(), "some")
+
 
 def test_assure_iter():
     s = {1}
-    assert assure_iter(None, set) == set()
-    assert assure_iter(1, set) == s
-    assert assure_iter(1, list) == [1]
-    assert assure_iter(s, set) is s
-    assert assure_iter(s, set, copy=True) is not s
+    assert ensure_iter(None, set) == set()
+    assert ensure_iter(1, set) == s
+    assert ensure_iter(1, list) == [1]
+    assert ensure_iter(s, set) is s
+    assert ensure_iter(s, set, copy=True) is not s
 
 
 def test_assure_list_copy():
     l = [1]
-    assert assure_list(l) is l
-    assert assure_list(l, copy=True) is not l
+    assert ensure_list(l) is l
+    assert ensure_list(l, copy=True) is not l
 
 
 def test_assure_list_from_str():
-    assert_equal(assure_list_from_str(''), None)
-    assert_equal(assure_list_from_str([]), None)
-    assert_equal(assure_list_from_str('somestring'), ['somestring'])
-    assert_equal(assure_list_from_str('some\nmultiline\nstring'), ['some', 'multiline', 'string'])
-    assert_equal(assure_list_from_str(['something']), ['something'])
-    assert_equal(assure_list_from_str(['a', 'listof', 'stuff']), ['a', 'listof', 'stuff'])
+    assert_equal(ensure_list_from_str(''), None)
+    assert_equal(ensure_list_from_str([]), None)
+    assert_equal(ensure_list_from_str('somestring'), ['somestring'])
+    assert_equal(ensure_list_from_str('some\nmultiline\nstring'), ['some', 'multiline', 'string'])
+    assert_equal(ensure_list_from_str(['something']), ['something'])
+    assert_equal(ensure_list_from_str(['a', 'listof', 'stuff']), ['a', 'listof', 'stuff'])
 
 
 def test_assure_dict_from_str():
-    assert_equal(assure_dict_from_str(''), None)
-    assert_equal(assure_dict_from_str({}), None)
+    assert_equal(ensure_dict_from_str(''), None)
+    assert_equal(ensure_dict_from_str({}), None)
     target_dict = dict(
         __ac_name='{user}', __ac_password='{password}',
         cookies_enabled='', submit='Log in'
     )
     string = '__ac_name={user}\n__ac_password={password}\nsubmit=Log ' \
                'in\ncookies_enabled='
-    assert_equal(assure_dict_from_str(string), target_dict)
-    assert_equal(assure_dict_from_str(
+    assert_equal(ensure_dict_from_str(string), target_dict)
+    assert_equal(ensure_dict_from_str(
         target_dict),
         target_dict)
 
@@ -502,8 +537,8 @@ def test_assure_bool():
         (['False', 0, '0', 'no', 'off'], False)
     ]:
         for v in values:
-            eq_(assure_bool(v), t)
-    assert_raises(ValueError, assure_bool, "unknown")
+            eq_(ensure_bool(v), t)
+    assert_raises(ValueError, ensure_bool, "unknown")
 
 
 def test_generate_chunks():
@@ -605,7 +640,10 @@ def test_expandpath():
     eq_(expandpath("some", False), expandvars('some'))
     assert_true(isabs(expandpath('some')))
     # this may have to go because of platform issues
-    eq_(expandpath("$HOME"), expanduser('~'))
+    if not on_windows:
+        # expanduser is not influenced by our HOME setting adjustments
+        # for the tests on windows
+        eq_(expandpath("$HOME"), expanduser('~'))
 
 
 def test_is_explicit_path():
@@ -666,7 +704,7 @@ def test_all_same():
         yield 'a'
         yield 'a'
         yield 'b'
-        raise ValueError("Should not get here since on b should return")
+        raise ValueError("Should not get here since on b should return")  # pragma: no cover
 
     ok_(not all_same(never_get_to_not_needed()))
 
@@ -707,9 +745,12 @@ def test_path_():
 def test_get_timestamp_suffix():
     # we need to patch temporarily TZ
     with patch.dict('os.environ', {'TZ': 'GMT'}):
+        # figure out how GMT time zone suffix is represented
+        # could be +0 or -0, depending on platform
+        # just use whatever it is, not the subject of this test
+        tz_suffix = time.strftime('%z', time.gmtime(0))
         # skynet DOB
-        target_ts = '1970-01-01T00:00:00-0000' \
-            if on_windows else '1970-01-01T00:00:00+0000'
+        target_ts = '1970-01-01T00:00:00' + tz_suffix
         assert_equal(get_timestamp_suffix(0), '-' + target_ts)
         assert_equal(get_timestamp_suffix(0, prefix="+"),
                      '+' + target_ts)
@@ -748,28 +789,28 @@ def test_memoized_generator():
 
 
 def test_assure_unicode():
-    ok_(isinstance(assure_unicode("m"), str))
-    ok_(isinstance(assure_unicode('grandchild_äöü東'), str))
-    ok_(isinstance(assure_unicode(u'grandchild_äöü東'), str))
-    eq_(assure_unicode('grandchild_äöü東'), u'grandchild_äöü東')
+    ok_(isinstance(ensure_unicode("m"), str))
+    ok_(isinstance(ensure_unicode('grandchild_äöü東'), str))
+    ok_(isinstance(ensure_unicode(u'grandchild_äöü東'), str))
+    eq_(ensure_unicode('grandchild_äöü東'), u'grandchild_äöü東')
     # now, non-utf8
     # Decoding could be deduced with high confidence when the string is
     # really encoded in that codepage
     mom_koi8r = u"мама".encode('koi8-r')
-    eq_(assure_unicode(mom_koi8r), u"мама")
-    eq_(assure_unicode(mom_koi8r, confidence=0.9), u"мама")
+    eq_(ensure_unicode(mom_koi8r), u"мама")
+    eq_(ensure_unicode(mom_koi8r, confidence=0.9), u"мама")
     mom_iso8859 = u'mamá'.encode('iso-8859-1')
-    eq_(assure_unicode(mom_iso8859), u'mamá')
-    eq_(assure_unicode(mom_iso8859, confidence=0.5), u'mamá')
+    eq_(ensure_unicode(mom_iso8859), u'mamá')
+    eq_(ensure_unicode(mom_iso8859, confidence=0.5), u'mamá')
     # but when we mix, it does still guess something allowing to decode:
     mixedin = mom_koi8r + u'東'.encode('iso2022_jp') + u'東'.encode('utf-8')
-    ok_(isinstance(assure_unicode(mixedin), str))
+    ok_(isinstance(ensure_unicode(mixedin), str))
     # but should fail if we request high confidence result:
     with assert_raises(ValueError):
-        assure_unicode(mixedin, confidence=0.9)
+        ensure_unicode(mixedin, confidence=0.9)
     # For other, non string values, actually just returns original value
     # TODO: RF to actually "assure" or fail??  For now hardcoding that assumption
-    assert assure_unicode(1) == 1
+    assert ensure_unicode(1) == 1
 
 
 def test_pathlib_unicode():
@@ -792,7 +833,7 @@ def test_as_unicode():
     assert_in("1 is not of any of known or provided", str(cme.exception))
 
 
-@known_failure_windows
+@skip_if_on_windows
 @with_tempfile(mkdir=True)
 def test_path_prefix(path):
     eq_(get_path_prefix('/d1/d2', '/d1/d2'), '')
@@ -860,7 +901,7 @@ def test_get_dataset_root(path):
         eq_(get_dataset_root(fname), os.curdir)
 
 
-@known_failure_windows
+@skip_if_on_windows
 def test_path_startswith():
     ok_(path_startswith('/a/b', '/a'))
     ok_(path_startswith('/a/b', '/a/b'))
@@ -876,7 +917,7 @@ def test_path_startswith():
     assert_raises(ValueError, path_startswith, '/a/b', 'a')
 
 
-@known_failure_windows
+@skip_if_on_windows
 def test_path_is_subpath():
     ok_(path_is_subpath('/a/b', '/a'))
     ok_(path_is_subpath('/a/b/c', '/a'))
@@ -893,22 +934,6 @@ def test_path_is_subpath():
     assert_raises(ValueError, path_is_subpath, '/a/b', 'a')
 
 
-def test_safe_print():
-    """Just to test that we are getting two attempts to print"""
-
-    called = [0]
-
-    def _print(s):
-        assert_equal(s, "bua")
-        called[0] += 1
-        if called[0] == 1:
-            raise UnicodeEncodeError('crap', u"", 0, 1, 'whatever')
-
-    with patch.object(builtins, 'print', _print):
-        safe_print("bua")
-    assert_equal(called[0], 2)
-
-
 def test_probe_known_failure():
 
     # Note: we can't test the switch "datalad.tests.knownfailures.probe"
@@ -997,32 +1022,6 @@ def test_known_failure():
         assert_raises(AssertionError, failing)
 
 
-def test_known_failure_v6():
-
-    @known_failure_v6
-    def failing():
-        raise AssertionError("Failed")
-
-    v6 = dl_cfg.obtain("datalad.repo.version") == 6
-    skip = dl_cfg.obtain("datalad.tests.knownfailures.skip")
-    probe = dl_cfg.obtain("datalad.tests.knownfailures.probe")
-
-    if v6:
-        if skip:
-            # skipping takes precedence over probing
-            failing()
-        elif probe:
-            # if we probe a known failure it's okay to fail:
-            failing()
-        else:
-            # not skipping and not probing results in the original failure:
-            assert_raises(AssertionError, failing)
-
-    else:
-        # behaves as if it wasn't decorated at all, no matter what
-        assert_raises(AssertionError, failing)
-
-
 from datalad.utils import read_csv_lines
 
 
@@ -1170,10 +1169,6 @@ def test_dlabspath(path):
             eq_(dlabspath("./bu", norm=True), opj(d, "bu"))
 
 
-# OSError: [WinError 998] Invalid access to memory location:
-# '(originated from NtWow64ReadVirtualMemory64)'
-# But passes on a real win10 box -- might be server config issue
-@known_failure_appveyor
 @with_tree({'1': 'content', 'd': {'2': 'more'}})
 def test_get_open_files(p):
     pobj = Path(p)
@@ -1195,7 +1190,7 @@ def test_get_open_files(p):
         # however, on windows get_open_files() can be very slow
         # (e.g. the first invocation in this test (above) can easily
         # take 30-50s). It is not worth slowing the tests to
-        # accomodate this issue, given we have tested proper functioning
+        # accommodate this issue, given we have tested proper functioning
         # in principle already above).
         return
 
@@ -1209,7 +1204,7 @@ def test_get_open_files(p):
                  stdout=PIPE,
                  cwd=str(subd))
     # Assure that it started and we read the OK
-    eq_(assure_unicode(proc.stdout.readline().strip()), u"OK")
+    eq_(ensure_unicode(proc.stdout.readline().strip()), u"OK")
     assert time() - t0 < 5 # that we were not stuck waiting for process to finish
     eq_(get_open_files(p)[str(subd.resolve())].pid, proc.pid)
     eq_(get_open_files(subd)[str(subd.resolve())].pid, proc.pid)
@@ -1253,12 +1248,16 @@ def test_create_tree(path):
             # right away an obscure case where we have both 1 and 1.gz
                 ('1', content*2),
                 ('1.gz', content*3),
+                ('1.xz', content*4),
+                ('1.lzma', content*5),
             ]
         )),
     ]))
     ok_file_has_content(op.join(path, '1'), content)
     ok_file_has_content(op.join(path, 'sd', '1'), content*2)
     ok_file_has_content(op.join(path, 'sd', '1.gz'), content*3, decompress=True)
+    ok_file_has_content(op.join(path, 'sd', '1.xz'), content*4, decompress=True)
+    ok_file_has_content(op.join(path, 'sd', '1.lzma'), content*5, decompress=True)
 
 
 def test_never_fail():
@@ -1287,12 +1286,22 @@ def test_never_fail():
 def test_is_interactive(fout):
     # must not fail if one of the streams is no longer open:
     # https://github.com/datalad/datalad/issues/3267
-    from ..cmd import Runner
+    from datalad.cmd import (
+        KillOutput,
+        NoCapture,
+        StdOutErrCapture,
+        WitlessRunner,
+    )
+    from datalad.support.gitrepo import GitProgress
+    from datalad.support.annexrepo import (
+        AnnexInitOutput,
+        AnnexJsonProtocol,
+    )
 
     bools = ["False", "True"]
 
     def get_interactive(py_pre="", **run_kwargs):
-        out, err = Runner().run(
+        out = WitlessRunner().run(
             [sys.executable,
              "-c",
              py_pre +
@@ -1309,9 +1318,121 @@ def test_is_interactive(fout):
         assert_in(out, bools)
         return bool(bools.index(out))
 
-    # we never request for pty in our Runner, so can't be interactive
-    eq_(get_interactive(), False)
+    # verify that NoCapture can make fully interactive execution
+    # happen, also test the core protocols
+    # (we can only be interactive in a runner, if the test execution
+    # itself happens in an interactive environment)
+    for proto, interactive in ((NoCapture,
+                                # It is unclear why (on travis only) a child
+                                # process can report to be interactive
+                                # whenever the parent process is not.
+                                # Maintain this test exception until
+                                # someone can provide insight. The point of
+                                # this test is to ensure that NoCapture
+                                # in an interactive parent also keeps the
+                                # child interactive, so this oddity is not
+                                # relevant.
+                                True if on_travis else is_interactive()),
+                               (KillOutput, False),
+                               (StdOutErrCapture, False),
+                               (GitProgress, False),
+                               (AnnexInitOutput, False),
+                               (AnnexJsonProtocol, False)):
+        eq_(get_interactive(protocol=proto),
+            interactive,
+            msg='{} -> {}'.format(str(proto), interactive))
     # and it must not crash if smth is closed
     for o in ('stderr', 'stdin', 'stdout'):
         eq_(get_interactive("import sys; sys.%s.close(); " % o), False)
 
+
+def test_splitjoin_cmdline():
+    # Do full round trip on a number of tricky samples
+    for args in (
+        ['cmd', '-o1', 'simple'],
+        ['c o', r'\m', ''],
+        ['c o', ' '],
+    ):
+        cmdline = join_cmdline(args)
+        assert isinstance(cmdline, str)
+        eq_(split_cmdline(cmdline), args)
+    # assure that there is no needless quoting
+    if on_windows:
+        # in quote_cmdlinearg we always quote on Windows
+        eq_(join_cmdline(['abc', 'def']), '"abc" "def"')
+    else:
+        eq_(join_cmdline(['abc', 'def']), 'abc def')
+
+
+@skip_if_root
+@with_tempfile
+def test_obtain_write_permission(path):
+    path = Path(path)
+
+    # there's nothing at path yet:
+    assert_raises(FileNotFoundError, obtain_write_permission, path)
+
+    # Revoke write permission
+    path.write_text("something")
+    path.chmod(path.stat().st_mode & ~stat.S_IWRITE)
+    assert_raises(PermissionError, path.write_text, "different thing")
+
+    # Obtain and try again:
+    obtain_write_permission(path)
+    path.write_text("different thing")
+
+    # Already having permission is no issue:
+    obtain_write_permission(path)
+    path.write_text("yet another thing")
+
+
+@skip_if_root
+@with_tempfile(mkdir=True)
+def test_ensure_write_permission(path):
+
+    # This is testing the usecase of write protected directories needed for
+    # messing with an annex object tree (as done by the ORA special remote).
+    # However, that doesn't work on Windows since we can't revoke write
+    # permissions for the owner of a directory (at least on VFAT - may be
+    # true for NTFS as well - don't know).
+    # Hence, on windows/crippledFS only test on a file.
+
+    dir_ = Path(path)
+    if not on_windows and has_symlink_capability:
+        # set up write-protected dir containing a file
+        file_ = dir_ / "somefile"
+        file_.write_text("whatever")
+        dir_.chmod(dir_.stat().st_mode & ~stat.S_IWRITE)
+        assert_raises(PermissionError, file_.unlink)
+
+        # contextmanager lets us do it and restores permissions afterwards:
+        mode_before = dir_.stat().st_mode
+        with ensure_write_permission(dir_):
+            file_.unlink()
+
+        mode_after = dir_.stat().st_mode
+        assert_equal(mode_before, mode_after)
+        assert_raises(PermissionError, file_.write_text, "new file can't be "
+                                                         "written")
+
+        assert_raises(FileNotFoundError, ensure_write_permission(dir_ /
+                      "non" / "existent").__enter__)
+
+        # deletion within context doesn't let mode restoration fail:
+        with ensure_write_permission(dir_):
+            dir_.rmdir()
+
+        dir_.mkdir()  # recreate, since next block is executed unconditionally
+
+    # set up write-protected file:
+    file2 = dir_ / "protected.txt"
+    file2.write_text("unchangeable")
+    file2.chmod(file2.stat().st_mode & ~stat.S_IWRITE)
+    assert_raises(PermissionError, file2.write_text, "modification")
+
+    # within context we can:
+    with ensure_write_permission(file2):
+        file2.write_text("modification")
+
+    # mode is restored afterwards:
+    assert_raises(PermissionError, file2.write_text, "modification2")
